@@ -1,247 +1,59 @@
 import datetime as dt
-import itertools
-import multiprocessing as mp
 import pickle
-import platform
-import select
 import socket
-import sys
-import time
 
 import pytest
 
 from mbedtls import hashlib
-from mbedtls.exceptions import TLSError
+from mbedtls._tls import _DTLSCookie as DTLSCookie
+from mbedtls._tls import _PSKSToreProxy as PSKStoreProxy
 from mbedtls.pk import RSA
 from mbedtls.tls import *
-from mbedtls.tls import _BaseConfiguration as BaseConfiguration
-from mbedtls.tls import _DTLSCookie as DTLSCookie
-from mbedtls.tls import _enable_debug_output
-from mbedtls.tls import _PSKSToreProxy as PSKStoreProxy
-from mbedtls.tls import _set_debug_level
-from mbedtls.tls import _TLSSession as TLSSession
+from mbedtls.tls import HandshakeStep, TLSSession
 from mbedtls.x509 import CRT, CSR, BasicConstraints
 
-try:
-    from contextlib import suppress
-except ImportError:
-    # Python 2.7
-    from contextlib2 import suppress
-
-
-try:
-    FileNotFoundError
-except NameError:
-    # Python 2.7
-    FileNotFoundError = OSError
-
-
-class Client:
-    def __init__(self, cli_conf, proto, srv_address, srv_hostname):
-        super().__init__()
-        self.cli_conf = cli_conf
-        self.proto = proto
-        self.srv_address = srv_address
-        self.srv_hostname = srv_hostname
-        self._sock = None
-
-    def __enter__(self):
-        self.start()
-        return self
-
-    def __exit__(self, *exc_info):
-        self.stop()
-
-    def __del__(self):
-        self.stop()
-
-    def do_handshake(self):
-        if not self._sock:
-            return
-
-        self._sock.do_handshake()
-
-    def echo(self, buffer, chunksize):
-        if not self._sock:
-            return
-
-        view = memoryview(buffer)
-        received = bytearray()
-        for idx in range(0, len(view), chunksize):
-            part = view[idx : idx + chunksize]
-            amt = self._sock.send(part)
-            received += self._sock.recv(2 << 13)
-        return received
-
-    def start(self):
-        if self._sock:
-            self.stop()
-
-        self._sock = ClientContext(self.cli_conf).wrap_socket(
-            socket.socket(socket.AF_INET, self.proto),
-            server_hostname=self.srv_hostname,
-        )
-        self._sock.connect(self.srv_address)
-
-    def stop(self):
-        if not self._sock:
-            return
-
-        with suppress(TLSError, OSError):
-            self._sock.close()
-        self._sock = None
-
-
-class Server:
-    def __init__(self, srv_conf, proto, conn_q):
-        super().__init__()
-        self.srv_conf = srv_conf
-        self.proto = proto
-        self.conn_q = conn_q
-        self._sock = None
-
-    def __enter__(self):
-        self.start()
-        return self
-
-    def __exit__(self, *exc_info):
-        self.stop()
-
-    def __del__(self):
-        self.stop()
-
-    def start(self):
-        if self._sock:
-            self.stop()
-
-        self._sock = ServerContext(self.srv_conf).wrap_socket(
-            socket.socket(socket.AF_INET, self.proto)
-        )
-        self._sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
-        self._sock.bind(
-            ("127.0.0.1" if platform.system() == "Windows" else "", 0)
-        )
-        if self.proto == socket.SOCK_STREAM:
-            self._sock.listen(1)
-        self.conn_q.put(self._sock.getsockname())
-
-    def stop(self):
-        if not self._sock:
-            return
-
-        with suppress(TLSError, OSError):
-            self._sock.close()
-        self._sock = None
-
-    def run(self, conn_handler):
-        with self:
-            {
-                TLSConfiguration: self._run_tls,
-                DTLSConfiguration: self._run_dtls,
-            }[type(self.srv_conf)](conn_handler)
-
-    def _run_tls(self, conn_handler):
-        assert self._sock
-        conn, addr = self._sock.accept()
-        try:
-            conn.do_handshake()
-        except TLSError:
-            conn.close()
-            return
-        try:
-            conn_handler(conn)
-        finally:
-            conn.close()
-
-    def _run_dtls(self, conn_handler):
-        assert self._sock
-        cli, addr = self._sock.accept()
-        cli.setcookieparam(addr[0].encode("ascii"))
-        with pytest.raises(HelloVerifyRequest):
-            cli.do_handshake()
-
-        _, (cli, addr) = cli, cli.accept()
-        _.close()
-        cli.setcookieparam(addr[0].encode("ascii"))
-        try:
-            cli.do_handshake()
-        except TLSError:
-            cli.close()
-            return
-        try:
-            conn_handler(cli)
-        finally:
-            cli.close()
-
-
-class EchoHandler:
-    def __init__(self, stop_ev, packet_size=4096):
-        self.stop_ev = stop_ev
-        self.packet_size = packet_size
-
-    def __call__(self, conn):
-        while not self.stop_ev.is_set():
-            readable, _, err = select.select([conn], [], [], 0.1)
-            if err:
-                break
-
-            for _ in readable:
-                # We use `send()` instead of `sendto()` for DTLS as well
-                # because the DTLS socket is connected.
-                received = conn.recv(self.packet_size)
-                sent = conn.send(received)
-
 
 class TestPickle:
-    @pytest.fixture
-    def session(self):
-        return TLSSession()
+    @pytest.mark.parametrize(
+        "obj", (TLSConfiguration(), DTLSConfiguration()), ids=type
+    )
+    def test_picklable(self, obj):
+        assert obj == pickle.loads(pickle.dumps(obj))
 
-    @pytest.fixture(params=[TLSConfiguration, DTLSConfiguration])
-    def conf(self, request):
-        return request.param()
-
-    @pytest.fixture(params=[ClientContext, ServerContext])
-    def context(self, request, conf):
-        return request.param(conf)
-
-    @pytest.fixture
-    def identity(self):
-        return lambda obj: pickle.loads(pickle.dumps(obj))
-
-    @pytest.fixture
-    def tls_wrapped_buffer(self, context):
-        return TLSWrappedBuffer(context)
-
-    @pytest.fixture
-    def tls_wrapped_socket(self, tls_wrapped_buffer):
-        return TLSWrappedSocket(socket.socket(), tls_wrapped_buffer)
-
-    def test_session(self, session):
+    @pytest.mark.parametrize(
+        "obj",
+        (
+            ClientContext(TLSConfiguration()),
+            ClientContext(DTLSConfiguration()),
+            ServerContext(TLSConfiguration()),
+            ServerContext(DTLSConfiguration()),
+            TLSSession(),
+            TLSWrappedBuffer(ClientContext(DTLSConfiguration())),
+            TLSWrappedBuffer(ClientContext(TLSConfiguration())),
+            TLSWrappedBuffer(ServerContext(DTLSConfiguration())),
+            TLSWrappedBuffer(ServerContext(TLSConfiguration())),
+            TLSWrappedSocket(
+                socket.socket(),
+                TLSWrappedBuffer(ClientContext(DTLSConfiguration())),
+            ),
+            TLSWrappedSocket(
+                socket.socket(),
+                TLSWrappedBuffer(ClientContext(TLSConfiguration())),
+            ),
+            TLSWrappedSocket(
+                socket.socket(),
+                TLSWrappedBuffer(ServerContext(DTLSConfiguration())),
+            ),
+            TLSWrappedSocket(
+                socket.socket(),
+                TLSWrappedBuffer(ServerContext(TLSConfiguration())),
+            ),
+        ),
+        ids=type,
+    )
+    def test_unpicklable(self, obj):
         with pytest.raises(TypeError) as excinfo:
-            pickle.dumps(session)
-
-        assert str(excinfo.value).startswith("cannot pickle")
-
-    def test_configuration(self, conf, identity):
-        assert conf == identity(conf)
-
-    def test_context(self, context, identity):
-        with pytest.raises(TypeError) as excinfo:
-            pickle.dumps(context)
-
-        assert str(excinfo.value).startswith("cannot pickle")
-
-    def test_tls_wrapped_buffer(self, tls_wrapped_buffer):
-        with pytest.raises(TypeError) as excinfo:
-            pickle.dumps(tls_wrapped_buffer)
-
-        assert str(excinfo.value).startswith("cannot pickle")
-
-    def test_tls_wrapped_socket(self, tls_wrapped_socket):
-        # Python socket.socket is not pickable.
-        with pytest.raises(TypeError) as excinfo:
-            pickle.dumps(tls_wrapped_socket)
+            pickle.dumps(obj)
 
         assert str(excinfo.value).startswith("cannot pickle")
 
@@ -356,6 +168,15 @@ class TestTLSRecordHeader:
         assert TLSRecordHeader.from_bytes(serialized) == header
 
 
+class TestTLSSession:
+    @pytest.fixture
+    def session(self):
+        return TLSSession()
+
+    def test_repr(self, session):
+        assert isinstance(repr(session), str)
+
+
 class Chain:
     @pytest.fixture(scope="class")
     def now(self):
@@ -608,6 +429,9 @@ class TestBaseContext:
         cls = request.param
         return cls(conf)
 
+    def test_repr(self, context):
+        assert isinstance(repr(context), str)
+
     def test_get_configuration(self, context, conf):
         assert conf
         assert context.configuration is conf
@@ -665,197 +489,112 @@ class TestServerContext(TestBaseContext):
         assert isinstance(context.wrap_buffers(), TLSWrappedBuffer)
 
 
-PSK_AUTHENTICATION_CIPHERS = (
-    "TLS-ECDHE-PSK-WITH-AES-256-CBC-SHA",
-    "TLS-ECDHE-PSK-WITH-AES-128-CBC-SHA",
-    "TLS-DHE-PSK-WITH-AES-256-CBC-SHA",
-    "TLS-DHE-PSK-WITH-AES-128-CBC-SHA",
-    "TLS-RSA-PSK-WITH-AES-256-CBC-SHA",
-    "TLS-RSA-PSK-WITH-AES-128-CBC-SHA",
-    "TLS-PSK-WITH-AES-256-CBC-SHA",
-    "TLS-PSK-WITH-AES-128-CBC-SHA",
-)
-
-
-def generate_configs(*configs):
-    for conf, versions in configs:
-        for version in versions:
-            yield conf, version
-
-
-class TestCommunication(Chain):
-    @pytest.fixture(
-        params=generate_configs(
-            (TLSConfiguration, TLSVersion), (DTLSConfiguration, DTLSVersion)
-        )
-    )
-    def configs(self, request):
-        return request.param
+class TestTLSWrappedBuffer:
+    @pytest.fixture
+    def server_hostname(self):
+        return "hostname"
 
     @pytest.fixture
-    def conf_cls(self, configs):
-        assert issubclass(configs[0], BaseConfiguration)
-        return configs[0]
+    def psk(self):
+        return ("cli", b"secret")
 
     @pytest.fixture
-    def version(self, configs):
-        assert isinstance(configs[1], (TLSVersion, DTLSVersion))
-        return configs[1]
-
-    @pytest.fixture
-    def version_min(self, conf_cls):
-        return {
-            TLSConfiguration: TLSVersion.MINIMUM_SUPPORTED,
-            DTLSConfiguration: DTLSVersion.MINIMUM_SUPPORTED,
-        }[conf_cls]
-
-    @pytest.fixture
-    def proto(self, conf_cls):
-        return {
-            TLSConfiguration: socket.SOCK_STREAM,
-            DTLSConfiguration: socket.SOCK_DGRAM,
-        }[conf_cls]
-
-    @pytest.fixture
-    def srv_conf(
-        self,
-        conf_cls,
-        version,
-        version_min,
-        trust_store,
-        certificate_chain,
-        srv_psk,
-        ciphers,
-    ):
-        return conf_cls(
-            trust_store=trust_store,
-            certificate_chain=certificate_chain,
-            lowest_supported_version=version_min,
-            highest_supported_version=version,
-            ciphers=ciphers,
-            pre_shared_key_store=srv_psk,
-            validate_certificates=False,
+    def cli_conf(self, psk):
+        return TLSConfiguration(
+            pre_shared_key=psk, validate_certificates=False
         )
 
     @pytest.fixture
-    def cli_conf(
-        self, conf_cls, version, version_min, trust_store, cli_psk, ciphers
-    ):
-        return conf_cls(
-            trust_store=trust_store,
-            lowest_supported_version=version_min,
-            highest_supported_version=version,
-            ciphers=ciphers,
-            pre_shared_key=cli_psk,
-            validate_certificates=True,
+    def srv_conf(self, psk):
+        return TLSConfiguration(
+            pre_shared_key_store=dict((psk,)), validate_certificates=False
         )
 
-    @pytest.fixture(params=[4])
-    def debug(self, srv_conf, cli_conf, request):
-        _enable_debug_output(srv_conf)
-        _enable_debug_output(cli_conf)
-        _set_debug_level(request.param)
-
-    @pytest.fixture(scope="class", params=[None])
-    def ciphers(self, request):
-        return request.param
-
-    @pytest.fixture(scope="class", params=["End Entity"])
-    def srv_hostname(self, request):
-        return request.param
-
-    @pytest.fixture(scope="class", params=[None])
-    def cli_psk(self, request):
-        return request.param
-
-    @pytest.fixture(scope="class", params=[None])
-    def srv_psk(self, request):
-        return request.param
-
-    @pytest.fixture(params=[False])
-    def buffer(self, request, randbytes):
-        return randbytes(5 * 16 * 1024)
-
-    @pytest.fixture(scope="class")
-    def trust_store(self, ca0_crt):
-        store = TrustStore()
-        store.add(ca0_crt)
-        return store
+    @pytest.fixture
+    def client(Self, cli_conf, server_hostname):
+        ctx = ClientContext(cli_conf).wrap_buffers(server_hostname)
+        yield ctx
+        ctx.shutdown()
 
     @pytest.fixture
-    def server(self, srv_conf, version, proto):
-        conn_q = mp.SimpleQueue()
-        stop_ev = mp.Event()
-        srv = Server(srv_conf, proto, conn_q)
-        runner = mp.Process(target=srv.run, args=(EchoHandler(stop_ev),))
+    def server(self, srv_conf, server_hostname):
+        ctx = ServerContext(srv_conf).wrap_buffers()
+        yield ctx
+        ctx.shutdown()
 
-        runner.start()
-        yield conn_q.get()
-        stop_ev.set()
-        runner.join()
+    def test_e2e_handshake_and_communicate(self, client, server):
+        def do_io(*, src, dst, amt=1024):
+            in_transit = src.peek_outgoing(amt)
+            src.consume_outgoing(len(in_transit))
+            dst.receive_from_network(in_transit)
 
-    @pytest.fixture
-    def client(self, server, srv_hostname, cli_conf, proto):
-        with Client(cli_conf, proto, server, srv_hostname) as client:
-            yield client
+        def do_handshake(*end_state_pair):
+            for end, state in end_state_pair:
+                end.do_handshake()
+                assert end.context._state is state
 
-    @pytest.mark.timeout(60)
-    @pytest.mark.usefixtures("server")
-    @pytest.mark.parametrize(
-        "srv_hostname", ["Wrong End Entity"], indirect=True
-    )
-    def test_host_name_verification_failure(self, client, srv_hostname):
-        with pytest.raises(TLSError):
-            client.do_handshake()
+        do_handshake(
+            (server, HandshakeStep.CLIENT_HELLO),
+            (client, HandshakeStep.CLIENT_HELLO),
+            (client, HandshakeStep.SERVER_HELLO),
+        )
 
-    @pytest.mark.timeout(60)
-    @pytest.mark.usefixtures("server")
-    @pytest.mark.parametrize(
-        "ciphers", [PSK_AUTHENTICATION_CIPHERS], indirect=True
-    )
-    @pytest.mark.parametrize(
-        "srv_psk", [{"client": b"the secret key"}], indirect=True
-    )
-    @pytest.mark.parametrize(
-        "cli_psk", [("client", b"the secret key")], indirect=True
-    )
-    @pytest.mark.parametrize("chunksize", [1024])
-    def test_psk_authentication_success(self, client, buffer, chunksize):
-        client.do_handshake()
-        assert client.echo(buffer, chunksize) == buffer
+        do_io(src=client, dst=server)
+        do_handshake(
+            (server, HandshakeStep.SERVER_HELLO),
+            (server, HandshakeStep.SERVER_CERTIFICATE),
+            (server, HandshakeStep.SERVER_KEY_EXCHANGE),
+            (server, HandshakeStep.CERTIFICATE_REQUEST),
+            (server, HandshakeStep.SERVER_HELLO_DONE),
+            (server, HandshakeStep.CLIENT_CERTIFICATE),
+            (server, HandshakeStep.CLIENT_KEY_EXCHANGE),
+        )
+        assert client.negotiated_protocol() == server.negotiated_protocol()
 
-    @pytest.mark.timeout(60)
-    @pytest.mark.usefixtures("server")
-    @pytest.mark.parametrize(
-        "ciphers", [PSK_AUTHENTICATION_CIPHERS], indirect=True
-    )
-    @pytest.mark.parametrize(
-        "srv_psk",
-        [
-            {"client": b"another key"},
-            {"another client": b"the secret key"},
-            {"another client": b"another key"},
-        ],
-        indirect=True,
-    )
-    @pytest.mark.parametrize(
-        "cli_psk", [("client", b"the secret key")], indirect=True
-    )
-    def test_psk_authentication_failure(self, client):
-        with pytest.raises(TLSError):
-            client.do_handshake()
+        do_io(src=server, dst=client)
+        do_handshake(
+            (client, HandshakeStep.SERVER_CERTIFICATE),
+            (client, HandshakeStep.SERVER_KEY_EXCHANGE),
+            (client, HandshakeStep.CERTIFICATE_REQUEST),
+            (client, HandshakeStep.SERVER_HELLO_DONE),
+            (client, HandshakeStep.CLIENT_CERTIFICATE),
+            (client, HandshakeStep.CLIENT_KEY_EXCHANGE),
+            (client, HandshakeStep.CERTIFICATE_VERIFY),
+            (client, HandshakeStep.CLIENT_CHANGE_CIPHER_SPEC),
+            (client, HandshakeStep.CLIENT_FINISHED),
+            (client, HandshakeStep.SERVER_CHANGE_CIPHER_SPEC),
+        )
+        assert (
+            client.negotiated_tls_version() == server.negotiated_tls_version()
+        )
 
-    @pytest.mark.timeout(60)
-    @pytest.mark.usefixtures("server")
-    @pytest.mark.parametrize("ciphers", (ciphers_available(),), indirect=True)
-    @pytest.mark.parametrize("chunksize", [1024])
-    def test_client_server(self, client, buffer, chunksize):
-        while True:
-            try:
-                client.do_handshake()
-            except (WantReadError, WantWriteError):
-                pass
-            else:
-                break
+        do_io(src=client, dst=server)
+        do_handshake(
+            (server, HandshakeStep.CERTIFICATE_VERIFY),
+            (server, HandshakeStep.CLIENT_CHANGE_CIPHER_SPEC),
+            (server, HandshakeStep.CLIENT_FINISHED),
+            (server, HandshakeStep.SERVER_CHANGE_CIPHER_SPEC),
+            (server, HandshakeStep.SERVER_FINISHED),
+            (server, HandshakeStep.FLUSH_BUFFERS),
+            (server, HandshakeStep.HANDSHAKE_WRAPUP),
+            (server, HandshakeStep.HANDSHAKE_OVER),
+        )
 
-        assert client.echo(buffer, chunksize) == buffer
+        do_io(src=server, dst=client)
+        do_handshake(
+            (client, HandshakeStep.SERVER_FINISHED),
+            (client, HandshakeStep.FLUSH_BUFFERS),
+            (client, HandshakeStep.HANDSHAKE_WRAPUP),
+            (client, HandshakeStep.HANDSHAKE_OVER),
+        )
+        assert client.cipher() == server.cipher()
+
+        secret = b"a very secret message"
+
+        amt = client.write(secret)
+        do_io(src=client, dst=server)
+        assert server.read(amt) == secret
+
+        amt = server.write(secret)
+        do_io(src=server, dst=client)
+        assert client.read(amt) == secret
